
package com.sunlabs.archsim.engine;

import com.sunlabs.archsim.globals.GlobalDefaults;
import com.sunlabs.archsim.globals.Globals;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * Common denominator for journalinig simulation results
 * @author ib27688
 */
abstract class Journal {
	
	/** Global settings */
	protected Globals globals= Globals.getInstance();
	
	/** record the data captured by an Object */
	void record(Object data) {
		Boolean globalLog= globals.booleanValue(GlobalDefaults.GLOBAL_LOG);
		boolean log= (globalLog != null)?
					globalLog.booleanValue()
				:	GlobalDefaults.GLOBAL_LOG_DEF;
		if (log) {		
			System.out.println(data);
		}
	}
	
	/** record an error */
	void error(Object data) {
		System.err.println(data);
	}
	
	/** record informational output */
	void info(Object data) {
		Boolean globalLog= globals.booleanValue(GlobalDefaults.GLOBAL_LOG);
		boolean log= (globalLog != null)?
					globalLog.booleanValue()
				:	GlobalDefaults.GLOBAL_LOG_DEF;
		if (log) {		
			System.out.println(data);
		}
	}
	
	/** set the output for all journal, errors, and info */
	abstract void setOutput(String fileName);
	
	
	/** set the output for the journal */
	abstract void setJournalOutput(String fileName);
	
	/** set the destination for errors */
	abstract void setErrorOutput(String fileName);
	
	/** set the destination for the info output */
	abstract void setInfoOutput(String fileName);
	
	/** close the output */
	abstract void done();
	
} // class Journal
