package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.Parser.Table.*;
import edu.berkeley.sbp.Sequence.Position;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** implements Tomita's Graph Structured Stack */
class GSS {

    static int count = 0;
    static int shifts = 0;
    static int reductions = 0;
    int resets = 0;
    int waits = 0;
    
    public GSS() { }

    private Phase.Node[] reducing_list = null;

    // FIXME: right now, these are the performance bottleneck
    HashMapBag<Sequence,Phase.Waiting> waiting         = new HashMapBag<Sequence,Phase.Waiting>();
    HashMapBag<Integer,Sequence>       performed       = new HashMapBag<Integer,Sequence>();
    HashMapBag<Integer,Sequence>       lastperformed   = new HashMapBag<Integer,Sequence>();
    HashMapBag<Integer,Sequence>       expected        = new HashMapBag<Integer,Sequence>();
    
    /** FIXME */
    Forest.Many finalResult;

    /** corresponds to a positions <i>between tokens</i> the input stream; same as Tomita's U_i's */
    class Phase<Tok> implements Invokable<State, Forest, Phase<Tok>.Node>, IntegerMappable, GraphViz.ToGraphViz, Iterable<Phase.Node> {

        public Iterator<Phase.Node> iterator() { return hash.iterator(); }
        public void invoke(State st, Forest result, Node n) {
            shifts++;
            good |= next.newNode(n, result, st, false);
        }

        /** the token immediately after this phase */
        final Tok token;

        private final int pos;

        boolean reducing;
        private IntPairMap<Phase.Node> hash;  /* ALLOC */
        private boolean closed;
        private boolean good;
        private Phase next = null;
        private Phase prev;
        private Input.Location location;
        public final Parser parser;

        private Forest forest;

        public Phase(Phase prev, Parser parser, Phase previous, Tok token, Input.Location location, Forest forest) throws ParseFailed {
            this.prev = prev;
            this.forest = forest;
            this.parser = parser;
            this.pos = previous==null ? 0 : previous.pos+1;
            this.token = token;
            this.location = location;
            performed.clear();
            reset();
        }

        public void reset() throws ParseFailed {
            waiting.clear();
            expected.clear();
            lastperformed.clear();
            lastperformed.addAll(performed);
            performed.clear();
            hash = new IntPairMap<Phase.Node>();
            reset = false;
            good = false;
            closed = false;
            reducing = false;
            finalResult = null;
            if (prev != null) prev.shift(this, forest);
        }

      
        public boolean isDone() throws ParseFailed {
            if (token != null) return false;
            if (token==null && finalResult==null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected end of file\n"), token, hash.values()), getLocation());
            return true;
        }

        public Input.Location getLocation() { return location; }

        /** add a new node (merging with existing nodes if possible)
         *  @param parent             the parent of the new node
         *  @param result             the SPPF result corresponding to the new node
         *  @param state              the state that the new node is in
         *  @param fromEmptyReduction true iff this node is being created as a result of a reduction of length zero (see GRMLR paper)
         *  @param start              the earliest part of the input contributing to this node (used to make merging decisions)
         */
        public boolean newNode(Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            Node p = hash.get(state, parent==null?null:parent.phase());
            if (p != null)  return newNode2(p, parent, pending, state, fromEmptyReduction);
            else            return newNode3(parent, pending, state, fromEmptyReduction);
        }
        public void newNode(Node parent, Forest pending, State state, boolean fromEmptyReduction, Position reduction) {
            int pos = parent==null?0:parent.phase()==null?0:parent.phase().pos;
            Sequence owner = reduction==null ? null : reduction.owner();
            if (reduction!=null) {
                if (owner.hates!=null) {
                    for (Sequence s : performed.getAll(pos))
                        if (owner.hates.contains(s))
                            return;
                    for (Sequence s : lastperformed.getAll(pos))
                        if (owner.hates.contains(s)) {
                            //System.out.println("now expecting ["+pos+"] => " + s);
                            expected.add(pos, s);
                            return;
                        }
                }
                if (owner.needs != null)
                    for(Sequence s : owner.needs)
                        if (!performed.contains(pos, s)) {
                            waiting.add(s, new Waiting(parent, pending, state, fromEmptyReduction, reduction));
                            return;
                        }
                if (!performed.contains(pos, owner)) {
                    performed.add(pos, owner);
                    if (owner.hated != null)
                        for(Sequence seq : owner.hated)
                            if (performed.contains(pos, seq)) {
                                performed.remove(pos, seq);
                                reset = true;
                            }
                }
            }
            if (!owner.lame)
                newNode(parent, pending, state, fromEmptyReduction);
            if (reduction != null) {
                boolean redo = true;
                while(redo) {
                    redo = false;
                    for(Waiting w : waiting.getAll(owner)) {
                        if (w.parent==parent || (parent!=null&&w.parent!=null&&w.parent.phase()==parent.phase())) {
                            waiting.remove(owner, w);
                            w.perform();
                            redo = true;
                            break;
                        }
                    }
                }
            }
        }

        private boolean newNode2(Node p, Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            if (p.merge(parent, pending)) return true;
            p.parents().add(parent, true);
            if (p!=parent && !fromEmptyReduction && reducing) p.performReductions(parent);
            return true;
        }

        private boolean newNode3(Node parent, Forest pending, State state, boolean fromEmptyReduction) {
            do {
                if (token != null && state.canShift(token)) break;
                if (state.isAccepting()) break;
                if (token==null) break;
                if (!state.canReduce(token)) return false;
                //if (count > 1) break;
                //if (r.numPop == 0) break;
                //r.reduce(pending, parent, null, Phase.this, null);
                //return;
            } while(false);

            Node n = new Node(parent, pending, state);  // ALLOC
            if (reducing) {
                n.performEmptyReductions();
                if (!fromEmptyReduction) n.performReductions(parent);
            }
            return true;
        }

        /** perform all reduction operations */
        public void reduce() throws ParseFailed {
            try {
                reducing = true;
                if (reducing_list==null || reducing_list.length < hash.size())
                    reducing_list = new Phase.Node[hash.size() * 4];
                hash.toArray(reducing_list);
                int num = hash.size();
                for(int i=0; i<num; i++) {
                    Node n = reducing_list[i];
                    n.performEmptyReductions();
                    // INVARIANT: we never "see" a node until its parent-set is complete, modulo merges
                }
                for(int i=0; i<num; i++) {
                    Node n = reducing_list[i];
                    reducing_list[i] = null;
                    n.performReductions();
                }
                if (reset) {
                    reset = false;
                    resets++;
                    throw new Reset();
                }                
                for(int i : expected)
                    for(Sequence s : expected.getAll(i))
                        if (!performed.contains(i, s)) {
                            //System.out.println("resetting due to pos="+i+": " + s + " " + System.identityHashCode(s));
                            resets++;
                            throw new Reset();
                        }
            } catch (Reset r) {
                reset();
                reduce();
            }
            count = 0;
        }

        private boolean reset = false;
        class Reset extends RuntimeException { }

        /** perform all shift operations, adding promoted nodes to <tt>next</tt> */
        public void shift(Phase next, Forest result) throws ParseFailed {
            // this massively improves GC performance
            if (prev!=null && parser.helpgc) {
                prev.hash = null;
            }
            this.next = next;
            closed = true;
            Forest res = null;
            boolean ok = false;
            for(Phase.Node n : hash.values()) {
                if (token == null && n.state.isAccepting()) {
                    if (finalResult==null) finalResult = new Forest.Many();
                    for(Object f : n.results())
                        finalResult.merge((Forest)f);
                }
                if (token == null) continue;
                n.state.invokeShifts(token, this, result, n);
            }

            if (!good && token!=null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected character ")+" \'"+
                                                        ANSI.purple(StringUtil.escapify(token+"", "\\\'\r\n"))+
                                                        "\' encountered at "+
                                                        ANSI.green(getLocation())+"\n", token, hash.values()),
                                        getLocation());
            if (token==null && finalResult==null)
                throw new ParseFailed(ParseFailed.error(ANSI.red("unexpected end of file\n"), token, hash.values()),
                                        getLocation());
        }


        class Waiting {
            Node parent;
            Forest pending;
            State state;
            boolean fromEmptyReduction;
            Position reduction;
            public Waiting(Node parent, Forest pending, State state, boolean fromEmptyReduction, Position reduction) {
                waits++;
                this.parent = parent;
                this.pending = pending;
                this.state = state;
                this.fromEmptyReduction = fromEmptyReduction;
                this.reduction = reduction;
            }
            public void perform() {
                //System.out.println("performing: " + reduction.position);
                newNode(parent, pending, state, fromEmptyReduction, reduction);
            }
        }
       
        // Node /////////////////////////////////////////////////////////////////////////////////

        /** a node in the GSS */
        final class Node implements Invokable<Position, Node, Node>, IntegerMappable, GraphViz.ToGraphViz {
            public FastSet<Node> set = new FastSet<Node>();

           
            private boolean allqueued = false;

            /** what state this node is in */
            public final Parser.Table<Tok>.State<Tok> state;

            /** which Phase this Node belongs to (node that Node is also a non-static inner class of Phase) */
            public  Phase phase() { return Phase.this; }

            private HashSet<Forest.Many> resultMap = new HashSet<Forest.Many>();
            public Iterable<Forest.Many> results() { return resultMap; }
            public FastSet<Node> parents() { return set; }
            public boolean merge(Node parent, Forest result) {
                // FIXME: inefficient!
                for(Forest.Many f : results()) {
                    if (f.parents.contains(parent) /* UGLY: */ && f.parents.size()==1) {
                        f.merge(result);
                        return true;
                    }
                }
                Forest.Many f = new Forest.Many();
                f.parents.add(parent);
                f.merge(result);
                resultMap.add(f);
                set.add(parent, true);
                return false;
            }

            public void performReductions() {
                if (allqueued) return;
                allqueued = true;
                state.invokeReductions(token, this, this, null);
            }

            public void performReductions(Node n2) {
                if (!allqueued) performReductions();
                else            state.invokeReductions(token, this, this, n2);
            }

            public void performEmptyReductions() { state.invokeReductions(token, this, null, null); }
            public final void invoke(Position r, Node n, Node n2) {
                reductions++;
                if (n==null || n2==null || r.pos==0) {
                    if (r.pos==0) {
                        if (n==null) n = this;
                        else return;
                    }
                    if (n==null) return;
                    Forest[] holder = new Forest[r.pos];
                    if (r.pos==0) n.finish(r, r.zero(), n.phase());
                    else          n.reduce(r, r.pos-1,  n.phase(), null);
                } else {
                    if (r.pos<=0) throw new Error("called wrong form of reduce()");
                    int pos = r.pos-1;
                    n.reduce(r, pos, n.phase(), n2);
                }
            }

            public void reduce(Position r, int pos, Phase target, Node only) {
                Forest[] holder = r.holder;
                Forest old = holder[pos];

                HashSet<Forest> rr = new HashSet<Forest>();
                for(Forest result : results()) rr.add(result);
                for(Forest result : rr)
                    for(Node child : ((Forest.Many<?>)result).parents) {
                        if (only != null && child!=only) continue;
                        holder[pos] = result;
                        if (pos==0) child.finish(r, r.rewrite(child.phase().getLocation().createRegion(phase().getLocation())), target);
                        else        child.reduce(r, pos-1, target, null);
                    }

                holder[pos] = old;
            }

            public void finish(Position r, Forest result, Phase<Tok> target) {
                Parser.Table<Tok>.State<Tok> state0 = state.gotoSetNonTerminals.get(r.owner());
                if (result==null) throw new Error();
                if (state0!=null)
                    target.newNode(this, result, state0, r.pos<=0, r);
            }

            private Node(Node parent, Forest pending, State state) {
                this.state = state;
                this.merge(parent, pending);
                Phase start = parent==null ? null : parent.phase();
                if (parent != null) parents().add(parent, true);
                if (Phase.this.hash.get(state, start) != null) throw new Error("severe problem!");
                Phase.this.hash.put(state, start, this);
            }
            public int toInt() { return idx; }
            private final int idx = node_idx++;

            // GraphViz //////////////////////////////////////////////////////////////////////////////

            public GraphViz.Node toGraphViz(GraphViz gv) {
                if (gv.hasNode(this)) return gv.createNode(this);
                GraphViz.Node n = gv.createNode(this);
                n.label = ""+state.toStringx();
                n.shape = "rectangle";
                n.fill = "green";
                for(Forest result : results()) n.edge(result, "");
                for(Node parent : parents()) n.edge(parent, "");
                ((GraphViz.Group)phase().toGraphViz(gv)).add(n);
                return n;
            }
            public boolean isTransparent() { return false; }
            public boolean isHidden() { return false; }

        }
        private int node_idx = 0;

        public int toInt() { return pos+1; }
        public int size() { return hash==null ? 0 : hash.size(); }

        // GraphViz //////////////////////////////////////////////////////////////////////////////

        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Group g = gv.createGroup(this);
            g.label = "Phase " + pos;
            g.color = "gray";
            g.cluster = true;
            return g;
        }
        public boolean isTransparent() { return false; }
        public boolean isHidden() { return false; }

    }
}
